package sun.nio.ch;

import java.nio.channels.ShutdownChannelGroupException;
import java.nio.channels.AsynchronousCloseException;
import java.net.InetSocketAddress;
import java.security.PrivilegedAction;
import java.security.AccessControlContext;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.NotYetBoundException;
import java.nio.channels.AcceptPendingException;
import java.security.AccessController;
import java.util.concurrent.Future;
import java.nio.channels.AsynchronousSocketChannel;
import java.nio.channels.CompletionHandler;
import java.io.IOException;
import java.util.concurrent.atomic.AtomicBoolean;
import sun.misc.Unsafe;

class WindowsAsynchronousServerSocketChannelImpl extends AsynchronousServerSocketChannelImpl implements Iocp.OverlappedChannel
{
    private static final Unsafe unsafe;
    private static final int DATA_BUFFER_SIZE = 88;
    private final long handle;
    private final int completionKey;
    private final Iocp iocp;
    private final PendingIoCache ioCache;
    private final long dataBuffer;
    private AtomicBoolean accepting;
    
    WindowsAsynchronousServerSocketChannelImpl(Iocp iocp) throws IOException {
        super(iocp);
        this.accepting = new AtomicBoolean();
        long handle = (long)IOUtil.fdVal(this.fd);
        int associate;
        try {
            associate = iocp.associate(this, handle);
        }
        catch (IOException ex) {
            closesocket0(handle);
            throw ex;
        }
        this.handle = handle;
        this.completionKey = associate;
        this.iocp = iocp;
        this.ioCache = new PendingIoCache();
        this.dataBuffer = WindowsAsynchronousServerSocketChannelImpl.unsafe.allocateMemory(88L);
    }
    
    @Override
    public <V, A> PendingFuture<V, A> getByOverlapped(long n) {
        return this.ioCache.remove(n);
    }
    
    @Override
    void implClose() throws IOException {
        closesocket0(this.handle);
        this.ioCache.close();
        this.iocp.disassociate(this.completionKey);
        WindowsAsynchronousServerSocketChannelImpl.unsafe.freeMemory(this.dataBuffer);
    }
    
    @Override
    public AsynchronousChannelGroupImpl group() {
        return this.iocp;
    }
    
    @Override
    Future<AsynchronousSocketChannel> implAccept(Object o, CompletionHandler<AsynchronousSocketChannel, Object> completionHandler) {
        if (!this.isOpen()) {
            ClosedChannelException ex3 = new ClosedChannelException();
            if (completionHandler == null) {
                return CompletedFuture.withFailure(ex3);
            }
            Invoker.invokeIndirectly(this, completionHandler, o, null, ex3);
            return null;
        }
        else {
            if (this.isAcceptKilled()) {
                throw new RuntimeException("Accept not allowed due to cancellation");
            }
            if (this.localAddress == null) {
                throw new NotYetBoundException();
            }
            WindowsAsynchronousSocketChannelImpl windowsAsynchronousSocketChannelImpl = null;
            IOException ex = null;
            try {
                this.begin();
                windowsAsynchronousSocketChannelImpl = new WindowsAsynchronousSocketChannelImpl(this.iocp, false);
            }
            catch (IOException ex2) {
                ex = ex2;
            }
            finally {
                this.end();
            }
            if (ex != null) {
                if (completionHandler == null) {
                    return CompletedFuture.withFailure(ex);
                }
                Invoker.invokeIndirectly(this, completionHandler, o, null, ex);
                return null;
            }
            else {
                AccessControlContext accessControlContext = (System.getSecurityManager() == null) ? null : AccessController.getContext();
                PendingFuture pendingFuture = new PendingFuture(this, completionHandler, o);
                WindowsAsynchronousServerSocketChannelImpl.AcceptTask context = new WindowsAsynchronousServerSocketChannelImpl.AcceptTask(windowsAsynchronousSocketChannelImpl, accessControlContext, pendingFuture);
                pendingFuture.setContext(context);
                if (!this.accepting.compareAndSet(false, true)) {
                    throw new AcceptPendingException();
                }
                if (Iocp.supportsThreadAgnosticIo()) {
                    context.run();
                    return pendingFuture;
                }
                Invoker.invokeOnThreadInThreadPool(this, context);
                return pendingFuture;
            }
        }
    }
    
    private static native void initIDs();
    
    private static native int accept0(long p0, long p1, long p2, long p3) throws IOException;
    
    private static native void updateAcceptContext(long p0, long p1) throws IOException;
    
    private static native void closesocket0(long p0) throws IOException;
    
    static void access$200(long n, long n2) throws IOException {
        updateAcceptContext(n, n2);
    }
    
    static int access$500(long n, long n2, long n3, long n4) throws IOException {
        return accept0(n, n2, n3, n4);
    }
    
    static {
        unsafe = Unsafe.getUnsafe();
        IOUtil.load();
        initIDs();
    }
    
    private class AcceptTask implements Runnable, Iocp.ResultHandler
    {
        private final WindowsAsynchronousSocketChannelImpl channel;
        private final AccessControlContext acc;
        private final PendingFuture<AsynchronousSocketChannel, Object> result;
        
        AcceptTask(WindowsAsynchronousSocketChannelImpl channel, AccessControlContext acc, PendingFuture<AsynchronousSocketChannel, Object> result) {
            this.channel = channel;
            this.acc = acc;
            this.result = result;
        }
        
        void enableAccept() {
            WindowsAsynchronousServerSocketChannelImpl.this.accepting.set(false);
        }
        
        void closeChildChannel() {
            try {
                this.channel.close();
            }
            catch (IOException ex) {}
        }
        
        void finishAccept() throws IOException {
            WindowsAsynchronousServerSocketChannelImpl.access$200(WindowsAsynchronousServerSocketChannelImpl.this.handle, this.channel.handle());
            InetSocketAddress localAddress = Net.localAddress(this.channel.fd);
            InetSocketAddress remoteAddress = Net.remoteAddress(this.channel.fd);
            this.channel.setConnected(localAddress, remoteAddress);
            if (this.acc != null) {
                AccessController.doPrivileged(new PrivilegedAction<Void>() {
                    @Override
                    public Void run() {
                        System.getSecurityManager().checkAccept(remoteAddress.getAddress().getHostAddress(), remoteAddress.getPort());
                        return null;
                    }
                }, this.acc);
            }
        }
        
        @Override
        public void run() {
            long add = 0L;
            try {
                WindowsAsynchronousServerSocketChannelImpl.this.begin();
                try {
                    this.channel.begin();
                    synchronized (this.result) {
                        add = WindowsAsynchronousServerSocketChannelImpl.this.ioCache.add(this.result);
                        if (WindowsAsynchronousServerSocketChannelImpl.access$500(WindowsAsynchronousServerSocketChannelImpl.this.handle, this.channel.handle(), add, WindowsAsynchronousServerSocketChannelImpl.this.dataBuffer) == -2) {
                            this.channel.end();
                            return;
                        }
                        this.finishAccept();
                        this.enableAccept();
                        this.result.setResult(this.channel);
                    }
                }
                finally {
                    this.channel.end();
                }
            }
            catch (Throwable o) {
                if (add != 0L) {
                    WindowsAsynchronousServerSocketChannelImpl.this.ioCache.remove(add);
                }
                this.closeChildChannel();
                if ((o instanceof ClosedChannelException)) {
                    o = new AsynchronousCloseException();
                }
                if (!(o instanceof IOException) && !(o instanceof SecurityException)) {
                    o = new IOException(o);
                }
                this.enableAccept();
                this.result.setFailure(o);
            }
            finally {
                WindowsAsynchronousServerSocketChannelImpl.this.end();
            }
            if (this.result.isCancelled()) {
                this.closeChildChannel();
            }
            Invoker.invokeIndirectly(this.result);
        }
        
        @Override
        public void completed(int n, boolean b) {
            try {
                if (WindowsAsynchronousServerSocketChannelImpl.this.iocp.isShutdown()) {
                    throw new IOException(new ShutdownChannelGroupException());
                }
                try {
                    WindowsAsynchronousServerSocketChannelImpl.this.begin();
                    try {
                        this.channel.begin();
                        this.finishAccept();
                    }
                    finally {
                        this.channel.end();
                    }
                }
                finally {
                    WindowsAsynchronousServerSocketChannelImpl.this.end();
                }
                this.enableAccept();
                this.result.setResult(this.channel);
            }
            catch (Throwable o) {
                this.enableAccept();
                this.closeChildChannel();
                if ((o instanceof ClosedChannelException)) {
                    o = new AsynchronousCloseException();
                }
                if (!(o instanceof IOException) && !(o instanceof SecurityException)) {
                    o = new IOException(o);
                }
                this.result.setFailure(o);
            }
            if (this.result.isCancelled()) {
                this.closeChildChannel();
            }
            Invoker.invokeIndirectly(this.result);
        }
        
        @Override
        public void failed(int n, IOException failure) {
            this.enableAccept();
            this.closeChildChannel();
            if (WindowsAsynchronousServerSocketChannelImpl.this.isOpen()) {
                this.result.setFailure(failure);
            }
            else {
                this.result.setFailure(new AsynchronousCloseException());
            }
            Invoker.invokeIndirectly(this.result);
        }
    }
}
